import { useMemo, useReducer, useEffect } from 'react';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

var byTextAscending = function byTextAscending(getTextProperty) {
  return function (objectA, objectB) {
    var upperA = getTextProperty(objectA).toUpperCase();
    var upperB = getTextProperty(objectB).toUpperCase();

    if (upperA < upperB) {
      return -1;
    }

    if (upperA > upperB) {
      return 1;
    }

    return 0;
  };
};
var byTextDescending = function byTextDescending(getTextProperty) {
  return function (objectA, objectB) {
    var upperA = getTextProperty(objectA).toUpperCase();
    var upperB = getTextProperty(objectB).toUpperCase();

    if (upperA > upperB) {
      return -1;
    }

    if (upperA < upperB) {
      return 1;
    }

    return 0;
  };
};

var createReducer = function createReducer() {
  return function (state, action) {
    switch (action.type) {
      case 'TOGGLE_SORT':
        if (!(action.columnName in state.columnsByName)) {
          throw new Error("Invalid column, " + action.columnName + " not found");
        }

        var isAscending = null;
        var sortedRows = []; // loop through all columns and set the sort parameter to off unless
        // it's the specified column (only one column at a time for )

        var columnCopy = state.columns.map(function (column) {
          // if the row was found
          if (action.columnName === column.name) {
            isAscending = column.sorted.asc;

            if (column.sort) {
              sortedRows = isAscending ? state.rows.sort(column.sort) : state.rows.sort(column.sort).reverse(); // default to sort by string
            } else {
              sortedRows = isAscending ? state.rows.sort(byTextAscending(function (object) {
                return object.original[action.columnName];
              })) : state.rows.sort(byTextDescending(function (object) {
                return object.original[action.columnName];
              }));
            }

            return _extends({}, column, {
              sorted: {
                on: true,
                asc: !column.sorted.asc
              }
            });
          }

          return _extends({}, column, {
            sorted: {
              on: false,
              asc: true
            }
          });
        });
        return _extends({}, state, {
          columns: columnCopy,
          rows: sortedRows,
          columnsByName: getColumnsByName(columnCopy)
        });

      case 'GLOBAL_FILTER':
        var filteredRows = action.filter(state.originalRows);
        var selectedRowsById = {};
        state.selectedRows.map(function (row) {
          selectedRowsById[row.id] = row.selected ? true : false;
        });
        return _extends({}, state, {
          rows: filteredRows.map(function (row) {
            return selectedRowsById[row.id] ? _extends({}, row, {
              selected: selectedRowsById[row.id]
            }) : _extends({}, row);
          }),
          filterOn: true
        });

      case 'GLOBAL_FILTER_OFF':
        var ogRows = [].concat(state.originalRows);
        var selRows = {};
        state.selectedRows.map(function (row) {
          selRows[row.id] = row.selected ? true : false;
        });
        return _extends({}, state, {
          filterOn: false,
          toggleAllState: false,
          rows: ogRows.map(function (row) {
            return selRows[row.id] ? _extends({}, row, {
              selected: selectedRowsById[row.id]
            }) : _extends({}, row);
          })
        });

      case 'SELECT_ROW':
        var stateCopy = _extends({}, state);

        stateCopy.rows = stateCopy.rows.map(function (row) {
          var newRow = _extends({}, row);

          if (newRow.id === action.rowId) {
            newRow.selected = !newRow.selected;
          }

          return newRow;
        });
        stateCopy.originalRows = stateCopy.originalRows.map(function (row) {
          var newRow = _extends({}, row);

          if (newRow.id === action.rowId) {
            newRow.selected = !newRow.selected;
          }

          return newRow;
        });
        stateCopy.selectedRows = stateCopy.originalRows.filter(function (row) {
          return row.selected === true;
        });
        stateCopy.toggleAllState = stateCopy.selectedRows.length === stateCopy.rows.length ? stateCopy.toggleAllState = true : stateCopy.toggleAllState = false;
        return stateCopy;

      case 'SEARCH_STRING':
        var stateCopySearch = _extends({}, state);

        stateCopySearch.rows = stateCopySearch.originalRows.filter(function (row) {
          return row.cells.filter(function (cell) {
            if (cell.value.includes(action.searchString)) {
              return true;
            }

            return false;
          }).length > 0;
        });
        return stateCopySearch;

      case 'TOGGLE_ALL':
        var stateCopyToggle = _extends({}, state);

        var rowIds = {};

        if (state.selectedRows.length < state.rows.length) {
          stateCopyToggle.rows = stateCopyToggle.rows.map(function (row) {
            rowIds[row.id] = true;
            return _extends({}, row, {
              selected: true
            });
          });
          stateCopyToggle.toggleAllState = true;
        } else {
          stateCopyToggle.rows = stateCopyToggle.rows.map(function (row) {
            rowIds[row.id] = false;
            return _extends({}, row, {
              selected: false
            });
          });
          stateCopyToggle.toggleAllState = false;
        }

        stateCopyToggle.originalRows = stateCopyToggle.originalRows.map(function (row) {
          return row.id in rowIds ? _extends({}, row, {
            selected: rowIds[row.id]
          }) : _extends({}, row);
        });
        stateCopyToggle.selectedRows = stateCopyToggle.originalRows.filter(function (row) {
          return row.selected;
        });
        return stateCopyToggle;

      default:
        throw new Error('Invalid reducer action');
    }
  };
};

var useTable = function useTable(columns, data, options) {
  var columnsWithSorting = useMemo(function () {
    return columns.map(function (column) {
      return _extends({}, column, {
        label: column.label ? column.label : column.name,
        hidden: column.hidden ? column.hidden : false,
        sort: column.sort,
        sorted: {
          on: false,
          asc: true
        }
      });
    });
  }, [columns]);
  var columnsByName = useMemo(function () {
    return getColumnsByName(columnsWithSorting);
  }, [columnsWithSorting]);
  var tableData = useMemo(function () {
    var sortedData = sortDataInOrder(data, columnsWithSorting);
    var newData = sortedData.map(function (row, idx) {
      return {
        id: idx,
        selected: false,
        hidden: false,
        original: row,
        cells: Object.entries(row).map(function (_ref) {
          var column = _ref[0],
              value = _ref[1];
          return {
            hidden: columnsByName[column].hidden,
            field: column,
            value: value,
            render: makeRender(value, columnsByName[column].render, row)
          };
        }).filter(function (cell) {
          return !cell.hidden;
        })
      };
    });
    return newData;
  }, [data, columnsWithSorting, columnsByName]);
  var reducer = createReducer();

  var _useReducer = useReducer(reducer, {
    columns: columnsWithSorting,
    columnsByName: columnsByName,
    originalRows: tableData,
    rows: tableData,
    selectedRows: [],
    toggleAllState: false,
    filterOn: false
  }),
      state = _useReducer[0],
      dispatch = _useReducer[1];

  var headers = useMemo(function () {
    return [].concat(state.columns.map(function (column) {
      var label = column.label ? column.label : column.name;
      return _extends({}, column, {
        render: makeHeaderRender(label, column.headerRender)
      });
    }));
  }, [state.columns]);
  useEffect(function () {
    if (options && options.filter) {
      dispatch({
        type: 'GLOBAL_FILTER',
        filter: options.filter
      });
    } else if (options && !options.filter) {
      dispatch({
        type: 'GLOBAL_FILTER_OFF'
      });
    }
  }, [options === null || options === void 0 ? void 0 : options.filter]);
  return {
    headers: headers.filter(function (column) {
      return !column.hidden;
    }),
    rows: state.rows,
    originalRows: state.originalRows,
    selectedRows: state.selectedRows,
    selectRow: function selectRow(rowId) {
      return dispatch({
        type: 'SELECT_ROW',
        rowId: rowId
      });
    },
    toggleAll: function toggleAll() {
      return dispatch({
        type: 'TOGGLE_ALL'
      });
    },
    toggleSort: function toggleSort(columnName) {
      return dispatch({
        type: 'TOGGLE_SORT',
        columnName: columnName
      });
    },
    setSearchString: function setSearchString(searchString) {
      return dispatch({
        type: 'SEARCH_STRING',
        searchString: searchString
      });
    },
    toggleAllState: state.toggleAllState
  };
};

var makeRender = function makeRender(value, render, row) {
  return render ? function () {
    return render({
      row: row,
      value: value
    });
  } : function () {
    return value;
  };
};

var makeHeaderRender = function makeHeaderRender(label, render) {
  return render ? function () {
    return render({
      label: label
    });
  } : function () {
    return label;
  };
};

var sortDataInOrder = function sortDataInOrder(data, columns) {
  return data.map(function (row) {
    var newRow = {};
    columns.forEach(function (column) {
      if (!(column.name in row)) {
        throw new Error("Invalid row data, " + column.name + " not found");
      }

      newRow[column.name] = row[column.name];
    });
    return newRow;
  });
};

var getColumnsByName = function getColumnsByName(columns) {
  var columnsByName = {};
  columns.forEach(function (column) {
    var col = {
      label: column.label
    };

    if (column.render) {
      col['render'] = column.render;
    }

    col['hidden'] = column.hidden;
    columnsByName[column.name] = col;
  });
  return columnsByName;
};

export { byTextAscending, byTextDescending, useTable };
//# sourceMappingURL=react-final-table.esm.js.map
